from os.path import exists
from numpy import ndarray
import numpy as np
from .datafolder import DataFolder

def import_dataset(name:str, data:str, force_download:bool=False, force_generate_npy=False)->ndarray:

    """Finds the dataset with the given name for the given type of data.

    Parameters
    ----------

    name : str
        Name of the dataset to find.
    data : str
        Name of the type of data to find the dataset for (structure or DMS).
    force_download : bool
        Whether to force download the dataset from HuggingFace Hub. Defaults to False.

    Returns
    -------

    dict: {str: ndarray}
        Dictionary with the following keys: 'references', 'sequences', and 'structure' or 'DMS' depending on the data type.
    """
    if data == 'dms': data = 'DMS'
    assert data in ['structure', 'DMS'], "data must be either 'structure' or 'DMS'"

    # Get the data folder
    try:
        assert not force_download, "Force download from HuggingFace Hub"
        datafolder = DataFolder.from_local(name=name, generate_npy=force_generate_npy)
        source = 'local'
        print("Using local data")
    except AssertionError as e:
        try:
            print(e)
            datafolder = DataFolder.from_huggingface(name=name)
            source = 'huggingface'
            print("Using data from HuggingFace Hub")
        except:
            raise ValueError("Dataset not found on HuggingFace Hub")

    if not exists(datafolder.get_references_npy()) or not exists(datafolder.get_sequences_npy()) or source == 'huggingface' or force_generate_npy:
        datafolder.generate_npy()

    out = {
        'references': np.load(datafolder.get_references_npy(), allow_pickle=True),
        'sequences': np.load(datafolder.get_sequences_npy(), allow_pickle=True)
        }

    # Get the dataset
    if data == 'structure':
        if not exists(datafolder.get_base_pairs_npy()) or force_download:
            datafolder.generate_npy()
        out['structure'] = np.load(datafolder.get_base_pairs_npy(), allow_pickle=True)

    elif data == 'DMS':
        if not exists(datafolder.get_dms_npy()) or force_download:
            datafolder.generate_npy()
        out['DMS'] = np.load(datafolder.get_dms_npy(), allow_pickle=True)

    return out



import os, json
import numpy as np
from .rnastructure import RNAstructure
from .util import DreemUtils
import pandas as pd


class Ct:
    def parse(ct_file):
        """Parse a ct file and return the sequence and structure

        Args:
            ct_file (str): path to ct file

        Returns:
            (str,str,str): (reference, sequence, paired_bases)


        """
        with open(ct_file, 'r') as f:
            lines = f.readlines()

        paired_bases, sequence = [], ''
        for line in lines[1:]:
            if line.strip() == '':
                break
            utr5, base, _, _, utr3, _ = line.split()
            sequence += base
            if int(utr3) > int(utr5) and int(utr3) != 0:
                paired_bases.append([int(utr5)-1, int(utr3)-1])

        return Ct.get_reference_from_title(ct_file), sequence, paired_bases


    def parse_list(ct_files):
        """Parse a list of ct files and return the sequences and structures"""
        return [Ct.parse(ct_file) for ct_file in ct_files]

    def get_reference_from_title(ct_file):
        return os.path.basename(ct_file).split('.')[0]

    def predict_dms(ct_file):
        """Predict the dms of a ct file using RNAstructure"""
        rna = RNAstructure()
        return rna.predictPairingProbability(Ct.parse(ct_file)[1])


class Fasta:
    def parse(fasta_file):
        """Parse a fasta file and return the references and sequences"""
        refs, seqs = [], []
        with open(fasta_file, 'r') as f:
            for line in f.readlines():
                if line[0] == '>':
                    refs.append(line[1:].strip())
                else:
                    seqs.append(line.strip())
        assert len(refs) == len(seqs), "The number of references and sequences in the fasta file must be the same ({} vs {})".format(len(refs), len(seqs))
        return seqs, refs

    def get_name(fasta_file):
        return os.path.basename(fasta_file).split('.')[0]

    def predict_structure(sequence):
        """Predict the structure of a sequence using RNAstructure"""
        rna = RNAstructure()
        return rna.predictStructure(sequence)

    def predict_dms(sequence):
        """Predict the dms of a sequence using RNAstructure"""
        rna = RNAstructure()
        return rna.predictPairingProbability(sequence)


class DreemOutput:
    def parse(dreem_output_file, drop_duplicates=True, max_mut=0.5):
        """Parse a dreem output file and return the references and sequences.

        Args:
            dreem_output_file (str): path to dreem output file
            drop_duplicates (bool): drop duplicate sequences
            max_mut (float): drop sequences with at least one mutation > max_mut

        Returns:
            (str,str,str): (reference, sequence, sub_rate)
        """

        df = pd.DataFrame(DreemUtils.flatten_json(DreemUtils.sort_dict(json.load(open(dreem_output_file, 'r')))))[['reference', 'sequence', 'sub_rate']]

        # df['max_mut'] = df.sub_rate.apply(lambda x: max(x))
        # len_before = len(df)
        # df = df[df.max_mut < max_mut].reset_index(drop=True)

        # len_before = len(df)
        # if drop_duplicates:
        #     df.drop_duplicates(subset='sequence', inplace=True)

        for _, row in df.iterrows():
            yield row['reference'], row['sequence'], row['sub_rate']




import os

# Define the one-hot encodings for the sequences and structures
seq2int = {
        'X': 0,
        'A': 1,
        'C': 2,
        'G': 3,
        'U': 4,
    }

int2seq = {v: k for k, v in seq2int.items()}

dot2int = {'.': 1, '(': 2, ')': 3, 'X': 0}
int2dot = {v: k for k, v in dot2int.items()}


def fastaToDict(fasta_file):
    """
    Reads a fasta file and returns a dictionary {sequence: reference}.

    Args:
        fasta_file (str): The path to the input fasta file.
    Returns:
        refSeq (dict): A dictionary with reference as key and sequence as value.


    Example:
    >>> import os
    >>> os.makedirs('temp', exist_ok=True)
    >>> f = open('temp/test.fasta', 'w')
    >>> f.write('>ref1\\nACGT\\n>ref2\\nCGTA')
    21
    >>> f.close()
    >>> print(fastaToDict('temp/test.fasta'))
    {'ref1': {'sequence': 'ACGT'}, 'ref2': {'sequence': 'CGTA'}}

    """

    refSeq = {}

    assert os.path.exists(fasta_file), f'The fasta file {fasta_file} does not exist.'
    with open(fasta_file, 'r') as handle:
        while True: # while not EOF
            ref, seq = handle.readline(), handle.readline()
            if ref == '' or seq == '':
                break
            assert ref[0] == '>', 'The reference should start with ">"'

            refSeq[ref[1:].strip()] = {'sequence': seq.strip().upper()}

    return refSeq


def add_braces_if_no_braces(s:str):
    if not s[0] == '{' or not s[-1] == '}':
        return '{' + s + '}'
    return s




class DreemUtils:
    def flatten_json(data):
        out, row = [], {}

        for k,v in data.copy().items():
            if type(v) != dict:
                row[k]= v
            else:
                row['reference'] = k
                for k2,v2 in v.items():
                    if type(v2) != dict:
                        row[k2] = v2
                    else:
                        row['section'] = k2
                        for k3,v3 in v2.items():
                            row['cluster'] = k3
                            if type(v3) != dict:
                                row[k3] = v3
                            else:
                                for k4,v4 in v3.items():
                                    row[k4] = v4
                                out.append(row.copy())
        return out

    def sort_dict(mut_profiles):
        sorting_key = lambda item: 1000*{int:0, str:0, float:0, list:1, dict:2}[type(item[1])] + ord(item[0][0])
        mut_profiles = {k:mut_profiles[k] for k in sorted(mut_profiles)}
        mut_profiles = dict(sorted(mut_profiles.items(), key=sorting_key))
        for k, v in mut_profiles.items():
            if type(v) is not dict:
                continue
            mut_profiles[k] = DreemUtils.sort_dict(v)
        return mut_profiles


def standardize_sequence(sequence):
    sequence = sequence.upper()
    sequence = sequence.replace('T', 'U')
    # trim the sequence
    sequence = sequence.replace(' ', '').replace('\n', '').replace('\t', '')
    return sequence

def sequence_has_regular_characters(sequence):
    return not (set(sequence) - set('ACGU'))



from .parsers import *
from .util import add_braces_if_no_braces, dot2int, int2dot, seq2int, standardize_sequence, sequence_has_regular_characters
import numpy as np
from .rnastructure import RNAstructure_singleton
from .env import env
class Datapoint:

    """A datapoint is a data structure where:
    - reference is the name of the sequence in the fasta file
    - sequence is a string of A, C, G, U
    - structure is a string of (, ), .
    - dms is a list of floats corresponding to the reactivity of each base in the sequence
    - paired_bases is a list of tuples (i,j) where i and j are paired bases.


    Example:
    >>> print(Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0]))
    "reference":{"sequence": "AACCGG", "paired_bases": [[0, 5], [1, 4]], "dms": [1.0, 2.0, 3.0]}
    >>> print(Datapoint(reference='reference', sequence='not a valid sequence', structure='((..))'))
    None
    >>> print(Datapoint(reference='reference', sequence='NNNNNNN', structure='((..))'))
    None
    >>> print(Datapoint(reference='reference', sequence='AACCGG', paired_bases=[(0, 5), (1, 4)]))
    "reference":{"sequence": "AACCGG", "paired_bases": [[0, 5], [1, 4]]}
    >>> print(Datapoint(reference='reference', sequence='AACCGG', paired_bases=[(1, 5), (1, 4)]))
    None
    """

    def __new__(cls, *args, **kwargs):
        # Check if the conditions are met before creating the object
        sequence = kwargs.get('sequence', args[0] if len(args) > 0 else None)
        reference = kwargs.get('reference', args[1] if len(args) > 1 else None)
        if sequence is None or reference is None:
            return None
        sequence = standardize_sequence(sequence)

        if not sequence_has_regular_characters(sequence) or len(sequence) == 0 or len(reference) == 0:
            return None
        
        if 'paired_bases' in kwargs:
            if not cls._assert_paired_bases(None, kwargs['paired_bases'], sequence):
                return None

        # If conditions are met, create and return the new instance
        instance = super().__new__(cls)
        return instance

    def __init__(self, sequence, reference, structure=None, dms=None, paired_bases=None):
        for attr in [sequence, reference]:
            assert isinstance(attr, str), f"Expected {attr} to be a string, got {type(attr)} instead."
            assert len(attr) > 0, f"Expected {attr} to be non-empty."

        # standardize the sequence
        sequence = standardize_sequence(sequence)

        if not sequence_has_regular_characters(sequence):
            raise Exception(f"Sequence {sequence} contains characters other than ACGTUacgtu.")

        self.reference = reference
        self.sequence = sequence
        self.structure = structure
        if paired_bases is not None or structure is not None:
            self.paired_bases = self._format_paired_bases(paired_bases) if paired_bases is not None else self._format_paired_bases(self.structure_to_paired_bases(structure))
        if dms is not None:
            self.dms = self._format_dms(dms) 
        self.opt_dict = {attr: eval(f'self.{attr}') for attr in ['paired_bases', 'dms'] if hasattr(self, attr)}

    def _format_paired_bases(self, paired_bases):
        """Returns a set of tuples.
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> datapoint._format_paired_bases([[1, 4], [0, 5]])
        ((0, 5), (1, 4))
        """
        return tuple(sorted([tuple(sorted(pair)) for pair in paired_bases])) if paired_bases is not None else None
    
    def _assert_paired_bases(self, paired_bases, sequence):
        """Ensures that the base pairs are 0=<bp<len(sequence) and that that there's maximum one pair per base. 
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> datapoint._assert_paired_bases([(0, 5), (1, 4)], 'AACCGG')
        True
        >>> datapoint._assert_paired_bases([(0, 5), (1, 4), (1, 2)], 'AACCGG')
        False
        >>> datapoint._assert_paired_bases([(0, 5), (4, 1), (1, 2)], 'AACCGG')
        False
        >>> datapoint._assert_paired_bases([(0, 8), (3, 4), (1, 2)], 'AACCGG')
        False
        >>> datapoint._assert_paired_bases(None, 'AACCGG')
        True
        """
        if paired_bases is not None:
            return all([0 <= pair[0] < len(sequence) and 0 <= pair[1] < len(sequence) and pair[0] != pair[1] for pair in paired_bases]) \
                and len(set([pair[0] for pair in paired_bases] + [pair[1] for pair in paired_bases])) == len(paired_bases) * 2
        else:
            return True

    def _format_dms(self, dms):
        """returns a tuple
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> datapoint._format_dms([1.0, 2.0, 3.0])
        (1.0, 2.0, 3.0)
        """
        if type(dms) == np.ndarray:
            dms = dms.tolist()
        return tuple(dms) 
        

    def to_dict(self):
        return self.reference, {'sequence': self.sequence, **{k:v for k,v in self.opt_dict.items() if v is not None}}

    def to_flat_dict(self):
        return {'reference': self.reference, 'sequence': self.sequence, **{k:v for k,v in self.opt_dict.items() if v is not None}}

    @classmethod
    def from_flat_dict(cls, d):
        return cls(sequence=d['sequence'], reference=d['reference'], **{k:v for k,v in d.items() if k not in ['sequence', 'reference']})

    def __str__(self):
        return '"'+self.reference+'"' + ':' + str({"sequence": self.sequence, **{k:v for k,v in self.opt_dict.items() if v is not None}}).replace("'",'"').replace('(','[').replace(')',']').replace('None','null')

    def __repr__(self) -> str:
        out = f"{self.__class__.__name__}('{self.reference}', sequence='{self.sequence}'"
        if hasattr(self, 'paired_bases'):
            out += f", paired_bases={self.paired_bases}"
        if hasattr(self, 'dms'):
            out += f", dms={self.dms}"
        return out + ')'

    def structure_to_paired_bases(self, structure):
        """Returns a list of tuples (i,j) where i and j are paired bases."""
        paired_bases = set()
        stack = []
        for i, char in enumerate(structure):
            if char == '(':
                stack.append(i)
            elif char == ')':
                paired_bases.add((stack.pop(), i))
        return paired_bases

    def embed_sequence(self):
        """Returns a list of integers corresponding to the sequence.

        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> datapoint.embed_sequence()
        array([1, 1, 2, 2, 3, 3])
        """
        return np.array([seq2int[base] for base in self.sequence])

    def embed_structure(self):
        """Returns a list of integers corresponding to the structure.

        >>> from numpy import array
        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> assert not (datapoint.embed_structure() - array([2, 2, 1, 1, 3, 3])).any(), 'The sequence is not embedded correctly.'
        """
        return np.array([dot2int[base] for base in self.structure])



class DatapointFactory:
    """A datapoint is a tuple (sequence, reference, structure, dms) where:

    - reference is the name of the sequence in the fasta file
    - sequence is a string of A, C, G, U
    - structure is a string of (, ), .
    - dms is a list of floats corresponding to the reactivity of each base in the sequence

    """
    def from_ct(ct_file, predict_dms):
        """Create a datapoint from a ct file. If predict_dms is True, the dms will be predicted using RNAstructure"""
        reference, sequence, paired_bases = Ct.parse(ct_file)
        sequence = standardize_sequence(sequence)

        if sequence_has_regular_characters(sequence):
            return Datapoint(sequence, reference, paired_bases=paired_bases, dms = Ct.predict_dms(ct_file) if predict_dms else None)

    def from_fasta(sequence, reference, predict_structure, predict_dms):
        """Create a datapoint from a fasta file. The structure and dms will be None."""
        sequence = standardize_sequence(sequence)

        if sequence_has_regular_characters(sequence):
            structure, dms = None, None
            if predict_structure:
                structure = Fasta.predict_structure(sequence)
            if predict_dms:
                dms = Fasta.predict_dms(sequence)
            return Datapoint(sequence, reference, structure, dms)

    def from_dreem_output(reference, sequence, mutation_rate, predict_structure):
        """Create a datapoint from a dreem output file. The structure and dms will be predicted if predict_structure and predict_dms are True."""
        sequence = standardize_sequence(sequence)
        if sequence_has_regular_characters(sequence):
            return Datapoint(
                sequence=sequence,
                reference=reference,
                structure=RNAstructure_singleton.predictStructure(sequence, dms=mutation_rate) if predict_structure else None,
                dms=mutation_rate)


    def from_json_line(string):
        """Create a datapoint from a json line. The json line should have the following format:
        "reference": {"sequence": "sequence", "paired_bases": [[1, 2], [3,4]], "dms": [1.0, 2.0, 3.0]}

        Example:
        >>> DatapointFactory.from_json_line('"reference": {"sequence": "ACAAGU"}')
        Datapoint('reference', sequence='ACAAGU')
        >>> DatapointFactory.from_json_line('"reference": {"sequence": "ACAAGU", "paired_bases": [[1, 2], [3, 4]], "dms": [1.0, 2.0, 3.0]}')
        Datapoint('reference', sequence='ACAAGU', paired_bases=((1, 2), (3, 4)), dms=(1.0, 2.0, 3.0))
        >>> DatapointFactory.from_json_line('"reference": {"sequence": "something else than ACGTUacgtu", "paired_bases": [[1, 2], [3, 4]], "dms": [1.0, 2.0, 3.0]}')
        """

        # process the string into a dictionary
        string= string.strip()[:-1] if string.strip()[-1] == ',' else string.strip()
        string = add_braces_if_no_braces(string)
        d = json.loads(string)
        ref = list(d.keys())[0]
        d = d[ref]

        # create the datapoint
        sequence = d['sequence']
        sequence = standardize_sequence(sequence)

        if sequence_has_regular_characters(sequence):
            return Datapoint(
                reference=ref,
                sequence=sequence,
                paired_bases=d['paired_bases'] if 'paired_bases' in d else None,
                dms=d['dms'] if 'dms' in d else None
            )







from .env import env
import os
from os.path import join

class PathDatafolder:
    """Path to files and folders of a datafolder of name `name`. The path to the data folder is `DATA_FOLDER`, which is defined in `env`.

    Parameters
    ----------

    name : str

    Returns
    -------

    PathDatafolder

    Example
    -------

    >>> path = PathDatafolder(name='my_test_datafolder_pytest')
    >>> path.name
    'my_test_datafolder_pytest'
    >>> print(path)
    PathDatafolder(name='my_test_datafolder_pytest')
    """

    def __init__(self, name, root = env.DATA_FOLDER) -> None:
        assert type(name) == str, f'name {name} is not a string'
        assert type(root) == str, f'root {root} is not a string'
        self.root = root
        self.name = name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name='{self.name}')"

    def get_data_folder(self)->str:
        """Returns the path to the data folder."""
        return self.root

    def get_main_folder(self)->str:
        """Returns the path to the main folder."""
        return join(self.get_data_folder(), self.name)

    def get_dms_npy(self)->str:
        """Returns the path to the DMS npy file"""
        return join(self.get_main_folder(), 'dms.npy')

    def get_base_pairs_npy(self)->str:
        """Returns the path to the structure npy file"""
        return join(self.get_main_folder(), 'base_pairs.npy')

    def get_references_npy(self)->str:
        """Returns the path to the names npy file"""
        return join(self.get_main_folder(), 'references.npy')

    def get_sequences_npy(self)->str:
        """Returns the path to the sequences npy file"""
        return join(self.get_main_folder(), 'sequences.npy')

    def get_json(self)->str:
        """Returns the path to the json."""
        return join(self.get_main_folder(), "data.json")

    def get_source_folder(self):
        """Returns the path to the source folder."""
        return join(self.get_main_folder(), "source")

    def get_source_files(self):
        """Returns the list of source files, if the source folder exists."""
        return [os.path.join(self.get_source_folder(), f) for f in os.listdir(self.get_source_folder())]

    def get_info_file(self):
        """Returns the path to the info file."""
        return join(self.get_main_folder(), "info.json")

    def get_readme_file(self):
        """Returns the path to the readme file."""
        return join(self.get_main_folder(), "README.md")



import os

def _source_env(path):
    """
    Source the environment variables from the file at path.

    Args:
        path (str): The path to the file to source.
    """
    out = {}
    with open(path, 'r') as f:
        for line in f.readlines():
            line = line.split('#')[0].strip()
            line = line.replace('export', '')
            key, value = line.split('=')
            key, value = key.replace(' ','').replace('"','').strip(), value.replace(' ','').replace('"','').strip()
            os.environ[key] = value
            out[key] = value

    return out


class Environment:
    
    def __init__(self, **kwargs):
        self.HUGGINGFACE_TOKEN = kwargs.get('HUGGINGFACE_TOKEN', os.environ.get('HUGGINGFACE_TOKEN', None)) 
        self.DATA_FOLDER = kwargs.get('DATA_FOLDER', os.environ.get('DATA_FOLDER', 'data/input_files'))
        self.DATA_FOLDER_TESTING = kwargs.get('DATA_FOLDER_TESTING', os.environ.get('DATA_FOLDER_TESTING', 'data/input_files_testing'))
        self.RNASTRUCTURE_PATH = kwargs.get('RNASTRUCTURE_PATH', os.environ.get('RNASTRUCTURE_PATH', ''))
        self.RNASTRUCTURE_TEMP_FOLDER = kwargs.get('RNASTRUCTURE_TEMP_FOLDER', os.environ.get('RNASTRUCTURE_TEMP_FOLDER', 'temp'))
        
    @classmethod
    def from_env(cls, path):
        return cls(**_source_env(path))
    
env = Environment()

def setup_env(path=None, **kwargs):
    """Setup the environment variables.
    
    Args:
        path (str): The path to the file to source.
        **kwargs: The environment variables to set.
            - HUGGINGFACE_TOKEN (str): The HuggingFace token.
            - DATA_FOLDER (str): The path to the data folder.
            - DATA_FOLDER_TESTING (str): The path to the data folder for testing.
            - RNASTRUCTURE_PATH (str): The path to the RNAstructure executable.
            - RNASTRUCTURE_TEMP_FOLDER (str): The path to the temporary folder for RNAstructure.

    Raises:
        ValueError: If an environment variable does not exist.
        
    Examples:
        >>> setup_env('env')
        >>> os.environ['RNASTRUCTURE_TEMP_FOLDER'].split('/')[-1]
        'temp'
    """
    
    # Source the environment variables from the file at path
    global env
    if path:
        env = Environment.from_env(path)

    # Update the environment variables with the kwargs
    for key, value in kwargs.items():
        if hasattr(env, key):
            setattr(env, key, value)
        else:
            raise ValueError(f'Environment variable {key} does not exist.')



import os
import numpy as np
from .datapoint import Datapoint, DatapointFactory
from typing import List
from .parsers import Fasta, DreemOutput
import pandas as pd
from tqdm import tqdm as tqdm_parser

class ListofDatapoints:

    def __init__(self, datapoints=[], verbose=True):
        self.datapoints, self.filtering_report = self.filter_duplicates_and_unvalid(datapoints, verbose)
        
    def __call__(self) -> List[Datapoint]:
        return self.datapoints

    @classmethod
    def from_fasta(cls, fasta_file, predict_structure, predict_dms, tqdm=True, verbose=True):
        """Create a list of datapoint from a fasta file. The structure and dms will be predicted if predict_structure and predict_dms are True."""
        sequences, references = Fasta.parse(fasta_file)
        return cls([DatapointFactory.from_fasta(sequence, reference, predict_structure, predict_dms)
                    for sequence, reference in tqdm_parser(zip(sequences, references), total=len(sequences), desc='Parsing fasta file', disable=not tqdm)], verbose=verbose)

    @classmethod
    def from_ct(cls, ct_files, predict_dms, tqdm=True, verbose=True):
        """Create a list of datapoint from a ct file. The dms will be predicted if predict_dms is True."""
        return cls([DatapointFactory.from_ct(ct_file, predict_dms) for ct_file in tqdm_parser(ct_files, total=len(ct_files), desc='Parsing ct files', disable=not tqdm)], verbose=verbose)

    @classmethod
    def from_dreem_output(cls, dreem_output_file, predict_structure, tqdm=True, verbose=True):
        """Create a list of datapoint from a dreem output file. The structure and dms will be predicted if predict_structure and predict_dms are True."""
        n_lines = len(list(DreemOutput.parse(dreem_output_file)))
        return cls([DatapointFactory.from_dreem_output(reference, sequence, mutation_rate, predict_structure)
            for reference, sequence, mutation_rate in tqdm_parser(DreemOutput.parse(dreem_output_file), total=n_lines, desc='Parsing dreem output file', disable=not tqdm)], verbose=verbose)

    @classmethod
    def from_json(cls, json_file, tqdm=True, verbose=True):
        """Create a list of datapoint from a json file."""
        with open(json_file) as f:
            return cls([DatapointFactory.from_json_line(line) for line in tqdm_parser(f, total=sum(1 for line in open(json_file) if line.strip() not in ['{', '}']), desc='Parsing json file', disable= not tqdm) if line.strip() not in ['{', '}']], verbose=verbose)


    def to_base_pairs_npy(self, path):
        """Writes the structure npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: structure matrix of pairs of 0-based indices.

        Examples:
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])], verbose=False)
            >>> datapoints.to_base_pairs_npy('temp/base_pairs.npy')
            array([[[1, 2],
                    [3, 4]]], dtype=object)
        """

        arr = np.array([np.array(datapoint.paired_bases) for datapoint in self.datapoints], dtype=object)
        np.save(path, arr)
        return arr


    def to_dms_npy(self, path):
        """Writes the dms npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: dms matrix of floats arrays.

        Examples:
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])], verbose=False)
            >>> datapoints.to_dms_npy('temp/dms.npy')
            array([[1.0, 2.0, 3.0]], dtype=object)
        """
        
        arr = np.array([np.array(datapoint.dms, dtype=np.float32) for datapoint in self.datapoints], dtype=object)
        np.save(path, arr, allow_pickle=True)
        return arr


    def to_sequence_npy(self, path):
        """Writes the sequence npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: int-encoded sequence matrix

        Examples:
            >>> from numpy import array
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0]),\
                                               Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])], verbose=False)
            >>> assert not (datapoints.to_sequence_npy('temp/sequence.npy') - array([[1, 1, 2, 2, 3, 3],[1, 1, 2, 2, 3, 3]], dtype=object)).any(), "The sequence matrix is not correct."
        """

        arr = np.array([datapoint.embed_sequence() for datapoint in self.datapoints], dtype=object)
        np.save(path, arr)
        return arr


    def to_reference_npy(self, path):
        """Writes the reference npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: reference matrix of int-encoded.

        Examples:
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])], verbose=False)
            >>> datapoints.to_reference_npy('temp/reference.npy')
            array(['reference'], dtype='<U9')
        """

        arr = np.array([datapoint.reference for datapoint in self.datapoints])
        np.save(path, arr)
        return arr


    def to_json(self, path) -> None:
        """Write a list of datapoints to a json file."""
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as f:
            f.write('{\n')
            for idx, datapoint in enumerate(self.datapoints): # write the datapoints one by one to avoid memory issues
                f.write(str(datapoint)+ ',\n' if idx != len(self.datapoints)-1 else str(datapoint)+'\n')
            f.write('}\n')

    def to_pandas(self, datapoints=None) -> pd.DataFrame:
        """Converts the list of datapoints into a pandas dataframe.

        Example:
        >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])], verbose=False)
        >>> datapoints.to_pandas()
           reference sequence      paired_bases              dms
        0  reference   AACCGG  ((1, 2), (3, 4))  (1.0, 2.0, 3.0)
        """
        if datapoints is None: datapoints = self.datapoints
        return pd.DataFrame([datapoint.to_flat_dict() for datapoint in datapoints])
    
    def from_pandas(self, df: pd.DataFrame) -> None:
        """Converts a pandas dataframe into a list of datapoints.

        Example:
            >>> df = pd.DataFrame([{'reference': 'reference', 'sequence': 'AACCGG', 'paired_bases': [[1, 2], [3, 4]], 'dms': [1.0, 2.0, 3.0]}])
            >>> ListofDatapoints(verbose=False).from_pandas(df)
            [Datapoint('reference', sequence='AACCGG', paired_bases=((1, 2), (3, 4)), dms=(1.0, 2.0, 3.0))]
        """
        return [Datapoint.from_flat_dict(datapoint_dict) for datapoint_dict in df.to_dict('records')]

    def filter_duplicates_and_unvalid(self, datapoints, verbose=True):
        """Filters out duplicate sequences.
        Only keep the first occurence of a sequence if all the other structures are the same.

        Examples:
            >>> datapoints = ListofDatapoints([ Datapoint(reference='ref1', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[0,0,0,0,0,0]),\
                                Datapoint(reference='ref2', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[0,0,0,0,0,0]),\
                                Datapoint(reference='ref2', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[0,0,0,0,0,0]),\
                                Datapoint(reference='ref4', sequence='AUGGC', paired_bases=[[1, 2]], dms=[0,0,0,0,1]),\
                                Datapoint(reference='ref5', sequence='AUGGC', paired_bases=[[0, 4], [2, 3]], dms=[0,0,0,0,0]),\
                                Datapoint(reference='ref6', sequence='not a regular sequence', paired_bases=[[0, 4]], dms=[0,0,0,0,0]) ])
            Over a total of 6 datapoints, there are:
                - 1 valid datapoints
                - 1 invalid datapoints (ex: sequence with non-regular characters)
                - 1 datapoints with the same reference
                - 1 duplicate sequences with the same structure / dms
                - 2 duplicate sequences with different structure / dms
            >>> datapoints.datapoints
            [Datapoint('ref1', sequence='AACCGG', paired_bases=((1, 2), (3, 4)), dms=(0, 0, 0, 0, 0, 0))]
        """
        
        # Remove None datapoints        
        n_input_datapoints = len(datapoints)
        datapoints = list(filter(lambda x: x is not None, datapoints))
        n_unvalid_datapoints = n_input_datapoints - len(datapoints)

        # Convert to pandas
        df = self.to_pandas(datapoints)
        
        # Remove duplicate or conflicting datapoints and keep track of the number of datapoints removed
        def drop_duplicates(df:pd.DataFrame, **kwargs):
            len_df_before = len(df)
            df.drop_duplicates(**kwargs)
            return len_df_before - len(df)
        
        # Don't allow multiple sequences with the same reference
        n_same_ref_datapoints = drop_duplicates(df, subset=['reference'], inplace=True)

        # Keep only one datapoint per sequence and structure
        if 'paired_bases' in df.columns:
            n_duplicates_datapoints = drop_duplicates(df, subset=['sequence','paired_bases'], inplace=True, ignore_index=True)

        # Keep only one datapoint per sequence and dms
        if 'dms' in df.columns:
            if not 'paired_bases' in df.columns: n_duplicates_datapoints = 0
            n_duplicates_datapoints += drop_duplicates(df, subset=['sequence','dms'], inplace=True, ignore_index=True)
        
        # If there are multiple structures / dms with the same structure, keep none
        n_same_seq_datapoints = drop_duplicates(df, subset=['sequence'], inplace=True, ignore_index=True, keep='first' if not ('paired_bases' in df.columns or 'dms' in df.columns) else False)

        # Convert back to list of datapoints
        datapoints = self.from_pandas(df)

        # Write report
        report = f"""Over a total of {n_input_datapoints} datapoints, there are:
    - {len(datapoints)} valid datapoints
    - {n_unvalid_datapoints} invalid datapoints (ex: sequence with non-regular characters)
    - {n_same_ref_datapoints} datapoints with the same reference"""
        if 'paired_bases' in df.columns or 'dms' in df.columns:
            report += f"""
    - {n_duplicates_datapoints} duplicate sequences with the same structure / dms
    - {n_same_seq_datapoints} duplicate sequences with different structure / dms"""
        else:
            report += f"""
    - {n_same_seq_datapoints} duplicate sequences"""
    
        if verbose: print(report)
            
        return datapoints, report
    
            
def base_pairs_into_matrix(base_pairs):
    """Turns a list of base pairs into a matrix of 1s and 0s.

    Args:
        base_pairs (list): list of base pairs, where each base pair is a tuple of 0-based two ints.

    Returns:
        matrix (np.array): matrix of 1s and 0s.

    Example:
        >>> from numpy import array, int64
        >>> base_pairs = [(0, 1), (2, 3), (4, 5)]
        >>> assert (base_pairs_into_matrix(base_pairs) ==  array([[0., 1., 0., 0., 0., 0.], \
                                                                [1., 0., 0., 0., 0., 0.], \
                                                                [0., 0., 0., 1., 0., 0.], \
                                                                [0., 0., 1., 0., 0., 0.], \
                                                                [0., 0., 0., 0., 0., 1.], \
                                                                [0., 0., 0., 0., 1., 0.]], dtype=int64)).all()

    """
    max_index = max([max(base_pair) for base_pair in base_pairs])
    arr = np.zeros((max_index + 1, max_index + 1))
    for base_pair in base_pairs:
        arr[base_pair[0], base_pair[1]] = 1
        arr[base_pair[1], base_pair[0]] = 1
    return arr


from setuptools import setup

setup(
    name="rouskinhf",
    packages=["rouskinhf"],
    license="MIT",
    install_requires=[
        'pytest',
        'pytest_cov',
        'numpy',
        'pyarrow',
        'pandas',
        'huggingface',
        'huggingface_hub',
        'jupyter',
        'ipykernel',
        'ipython',
        'scipy',
        'pydantic',
        'pytest-dotenv',
        'python-dotenv',
        'tqdm',
        'pytest-env'
        ],
)


import datetime, os, json
from .path import PathDatafolder

STRUCTURE_FROM_RNASTRUCTURE = 'RNAstructure'
STRUCTURE_FROM_SOURCE = 'source'
DMS_FROM_RNASTRUCTURE = 'RNAstructure'
DMS_FROM_SOURCE = 'source'

class InfoFileWriterTemplate(PathDatafolder):

    def __init__(self, name, root, has_dms, has_structure) -> None:
        super().__init__(name, root)
        self.has_dms = has_dms
        self.has_structure = has_structure

        self.info = {
            'name': self.name,
            'upload_date': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            'user': os.environ['USER'],
        }

    def write(self):
        """Write the info file and the readme file."""
        with open(self.get_info_file(), 'w') as f:
            json.dump(self.info, f, indent=4)

        with open(self.get_readme_file(), 'w') as f:
            f.write(f"""
---
license: mit
language:
  - en
tags:
  - chemistry
  - biology`
author: Silvi Rouskin
pretty_name: {self.name}
---

""")
            f.write(f"# {self.name}\n")
            for k, v in self.info.items():
                if k not in ['structure', 'about structure', 'DMS', 'about DMS','name', 'filtering_report']:
                    f.write(f"\t{k}: {v}\n")
            f.write(f"\n\tData types:")
            if self.has_structure:
                f.write(f"\n\t- structure ({self.info['about structure']})")
            if self.has_dms:
                f.write(f"\n\t- DMS ({self.info['about DMS']})")
            if 'filtering_report' in self.info:
                f.write(f"\n\n\tFiltering report: \n\t{self.info['filtering_report']}")
        return self
    
    def add_filtering_report(self, filtering_report):
        self.info['filtering_report'] = filtering_report
        return self


class InfoFileWriterFromDREEMoutput(InfoFileWriterTemplate):

    def __init__(self, name, root, path_in, predict_dms, predict_structure) -> None:
        super().__init__(name, root, has_dms=True, has_structure=predict_structure)

        file = json.load(open(path_in, 'r'))
        if 'user' in file:
            file['experimenter'] = file.pop('user')
        for k, v in file.items():
            if type(v) != dict:
                self.info[k] = v

        self.info['source'] = "`{}` (DREEM output format)".format(os.path.basename(path_in))
        if predict_structure:
            self.info['structure'] = STRUCTURE_FROM_RNASTRUCTURE
            self.info['about structure'] = 'Predicted using RNAstructure and the DMS reactivity data.'
        self.info['DMS'] = DMS_FROM_SOURCE
        self.info['about DMS'] = 'Measured experimentally using the attached experimental conditions.'


class InfoFileWriterFromCT(InfoFileWriterTemplate):

    def __init__(self, name, root, path_in, predict_dms, predict_structure) -> None:
        super().__init__(name, root, has_dms=predict_dms, has_structure=True)

        self.info['source'] = "`{}` (CT files format)".format(os.path.basename(path_in))
        self.info['structure'] = STRUCTURE_FROM_SOURCE
        self.info['about structure'] = 'Read from source.'
        if predict_dms:
            self.info['DMS'] = DMS_FROM_RNASTRUCTURE
            self.info['about DMS'] = 'Predicted using RNAstructure at 310K.'


class InfoFileWriterFromFasta(InfoFileWriterTemplate):

    def __init__(self, name, root, path_in, predict_dms, predict_structure) -> None:
        super().__init__(name, root, has_dms=predict_dms, has_structure=predict_structure)

        self.info['source'] = "`{}` (fasta file format)".format(os.path.basename(path_in))
        if predict_structure:
            self.info['structure'] = STRUCTURE_FROM_RNASTRUCTURE
            self.info['about structure'] = 'Predicted using RNAstructure at 310K.'
        if predict_dms:
            self.info['DMS'] = DMS_FROM_RNASTRUCTURE
            self.info['about DMS'] = 'Predicted using RNAstructure at 310K.'


def infoFileWriter(source, datafolder, predict_dms, predict_structure):

    if source == 'dreem_output':
        return InfoFileWriterFromDREEMoutput(datafolder.name, datafolder.path_out, datafolder.path_in, predict_dms=predict_dms, predict_structure=predict_structure)
    if source == 'ct':
        return InfoFileWriterFromCT(datafolder.name, datafolder.path_out, datafolder.path_in, predict_dms=predict_dms, predict_structure=predict_structure)
    if source == 'fasta':
        return InfoFileWriterFromFasta(datafolder.name, datafolder.path_out, datafolder.path_in, predict_dms=predict_dms, predict_structure=predict_structure)
    raise ValueError(f"Unknown source: {source}")

